// Copyright 2021 RnD Center "ELVEES", JSC

#include <errno.h>
#include <fcntl.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <sys/ioctl.h>
#include <unistd.h>

#include <linux/elcore50.h>

static void help(const char *program_name) {
    printf("ELcore control utility\n\n");
    printf("Usage: %s [COMMAND]\n\n", program_name);
    printf("Commands:\n");
    printf("    ps    Report a snapshot of the current ELcore jobs\n");
}

static int get_elcore_count() {
    int elcore_fd;
    struct elcore50_device_info info;

    elcore_fd = open("/dev/elcore0", O_RDWR, 0);
    if (elcore_fd < 0) return 0;

    if (ioctl(elcore_fd, ELCORE50_IOC_GET_CORE_IDX, &info) < 0) return 0;
    close(elcore_fd);
    return info.nclusters * info.cluster_cap;
}

static const char* state_to_str(enum elcore50_job_instance_state state) {
    switch (state) {
    case ELCORE50_JOB_STATUS_ENQUEUED:
        return "ENQUEUED";
    case ELCORE50_JOB_STATUS_RUN:
        return "RUNNING";
    case ELCORE50_JOB_STATUS_INTERRUPTED:
        return "INTERRUPTED";
    case ELCORE50_JOB_STATUS_SYSCALL:
        return "SYSCALL";
    case ELCORE50_JOB_STATUS_DONE:
        return "DONE";
    }
    return "UNKNOWN";
}

/* Mark job from shared job list if it also is in local job list due to possible
 * race condition between two ioctl's. */
static void shared_job_list_validation(struct elcore50_job_instance_list *shared_job_list,
                                       struct elcore50_job_instance_info local_info) {
    for (int i = 0; i < shared_job_list->job_instance_ret; ++i) {
        if (shared_job_list->info[i].pid == local_info.pid) {
            shared_job_list->info[i].pid = 0;
            return;
        }
     }
}

static int elcore_ps() {
    struct elcore50_job_instance_list list;
    struct elcore50_job_instance_list shared_job_list;
    char devname[32];
    int fd, ret;
    __u32 shared_job_count, job_count;

    fd = open("/dev/elcore", O_RDWR, 0);
    if (fd < 0) {
        fprintf(stderr, "Could not open device file /dev/elcore\n");
        return fd;
    }

    ret = ioctl(fd, ELCORE50_IOC_GET_JOB_COUNT, &shared_job_count);
    if (ret < 0) {
        fprintf(stderr, "ELCORE50_IOC_GET_JOB_COUNT failed with %d\n", errno);
        return ret;
    }

    if (shared_job_count) {
        shared_job_list.info = malloc(shared_job_count * sizeof(struct elcore50_job_instance_info));
        if (!shared_job_list.info) {
            fprintf(stderr, "Failed to allocate memory\n");
            return ret;
        }
        shared_job_list.job_instance_count = shared_job_count;
        shared_job_list.job_instance_ret = 0;

        ret = ioctl(fd, ELCORE50_IOC_GET_JOB_LIST, &shared_job_list);
        if (ret < 0) {
            free(shared_job_list.info);
            fprintf(stderr, "ELCORE50_IOC_GET_JOB_LIST failed with %d\n", errno);
            return ret;
        }
    }
    close(fd);

    printf("%-10s%-10s%-25s%-25s%-25s\n", "ID", "PID", "DEVICE", "STATE", "NAME");
    for (int dev = 0; dev < get_elcore_count(); ++dev) {
        sprintf(devname, "/dev/elcore%i", dev);
        if (access(devname, F_OK ) != 0)
            continue;  // file does not exist
        fd = open(devname, O_RDWR, 0);
        if (fd < 0) {
            if (shared_job_count) free(shared_job_list.info);
            fprintf(stderr, "Could not open device file %s\n", devname);
            return fd;
        }

        ret = ioctl(fd, ELCORE50_IOC_GET_JOB_COUNT, &job_count);
        if (ret < 0) {
            if (shared_job_count) free(shared_job_list.info);
            fprintf(stderr, "ELCORE50_IOC_GET_JOB_COUNT failed with %d\n", errno);
            return ret;
        }
        if (!job_count) continue;

        list.info = malloc(job_count * sizeof(struct elcore50_job_instance_info));
        if (!list.info) {
            if (shared_job_count) free(shared_job_list.info);
            fprintf(stderr, "Failed to allocate memory\n");
            return ret;
        }
        list.job_instance_count = job_count;
        list.job_instance_ret = 0;

        ret = ioctl(fd, ELCORE50_IOC_GET_JOB_LIST, &list);
        if (ret < 0) {
            free(list.info);
            if (shared_job_count) free(shared_job_list.info);
            fprintf(stderr, "ELCORE50_IOC_GET_JOB_LIST failed with %d\n", errno);
            return ret;
        }
        for (int i = 0; i < list.job_instance_ret; ++i) {
            struct elcore50_job_instance_info info = list.info[i];
            if ((info.state != ELCORE50_JOB_STATUS_ENQUEUED) && shared_job_count)
                shared_job_list_validation(&shared_job_list, info);
            printf("%-10lx%-10d%-25s%-25s%-25s\n", info.id, info.pid, devname,
                   state_to_str(info.state), info.name);
        }
        free(list.info);
        close(fd);
    }

    if (shared_job_count) {
        for (int i = 0; i < shared_job_list.job_instance_ret; ++i) {
                struct elcore50_job_instance_info info = shared_job_list.info[i];
                if (!info.pid) continue;
                printf("%-10lx%-10d%-25s%-25s%-25s\n", info.id, info.pid, "-",
                       state_to_str(info.state), info.name);
        }
        free(shared_job_list.info);
    }
    return 0;
}

int main(int argc, char **argv) {
    int opt;

    while ((opt = getopt(argc, argv, "h")) != -1) {
        switch (opt) {
        case 'h':
            help(argv[0]);
            return 0;
        default:
            fprintf(stderr, "Use %s -h for help.\n", argv[0]);
            return 1;
        }
    }

    if (optind >= argc) {
        fprintf(stderr, "Need command. Use %s -h for help.\n", argv[0]);
        return 1;
    }

    char *command = argv[optind];
    if (!strcmp(command, "ps")) {
        return elcore_ps();
    } else {
        fprintf(stderr, "Wrong command %s. Use %s -h for help.\n", command, argv[0]);
        return 1;
    }
}
